/*
 *
 *    OPEN-XCHANGE legal information
 *
 *    All intellectual property rights in the Software are protected by
 *    international copyright laws.
 *
 *
 *    In some countries OX, OX Open-Xchange, open xchange and OXtender
 *    as well as the corresponding Logos OX Open-Xchange and OX are registered
 *    trademarks of the OX Software GmbH. group of companies.
 *    The use of the Logos is not covered by the GNU General Public License.
 *    Instead, you are allowed to use these Logos according to the terms and
 *    conditions of the Creative Commons License, Version 2.5, Attribution,
 *    Non-commercial, ShareAlike, and the interpretation of the term
 *    Non-commercial applicable to the aforementioned license is published
 *    on the web site http://www.open-xchange.com/EN/legal/index.html.
 *
 *    Please make sure that third-party modules and libraries are used
 *    according to their respective licenses.
 *
 *    Any modifications to this package must retain all copyright notices
 *    of the original copyright holder(s) for the original code used.
 *
 *    After any such modifications, the original and derivative code shall remain
 *    under the copyright of the copyright holder(s) and/or original author(s)per
 *    the Attribution and Assignment Agreement that can be located at
 *    http://www.open-xchange.com/EN/developer/. The contributing author shall be
 *    given Attribution for the derivative code and a license granting use.
 *
 *     Copyright (C) 2016-2020 OX Software GmbH.
 *     Mail: info@open-xchange.com
 *
 *
 *     This program is free software; you can redistribute it and/or modify it
 *     under the terms of the GNU General Public License, Version 2 as published
 *     by the Free Software Foundation.
 *
 *     This program is distributed in the hope that it will be useful, but
 *     WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *     or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
 *     for more details.
 *
 *     You should have received a copy of the GNU General Public License along
 *     with this program; if not, write to the Free Software Foundation, Inc., 59
 *     Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

package com.openexchange.ant.data;

import java.io.File;
import java.io.FilenameFilter;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Stack;
import com.openexchange.ant.Tools;

/**
 * Sorts bundles according to their dependencies.
 *
 * @author <a href="mailto:marcus.klein@open-xchange.com">Marcus Klein</a>
 */
public class DependenciesSorter {

    public DependenciesSorter() {
        super();
    }

    /**
     * Visits all modules.
     *
     * @param classPathModules Base modules not to visit that can be used to solve dependencies.
     * @param appModules Modules to visit and to sort according to their dependencies.
     * @return a sorted list of modules ordered for the build.
     * @throws Exception if a cyclic dependency is detected.
     */
    public List<AbstractModule> sortDependencies(final List<AbstractModule> classPathModules, final List<AbstractModule> appModules) throws Exception {
        final List<AbstractModule> sorted = new LinkedList<AbstractModule>(classPathModules);
        // Find first projects containing only platform bundles. These should be built first.
        final List<AbstractModule> first = new LinkedList<AbstractModule>();
        for (final AbstractModule module : appModules) {
            if (!module.getDependencies().isEmpty()) {
                continue;
            }
            File jarsDir = new File(module.getFile(), "jars");
            if (!jarsDir.exists() || !jarsDir.isDirectory()) {
                continue;
            }
            String[] bundles = jarsDir.list(new FilenameFilter() {
                public boolean accept(File dir, String name) {
                    return name.endsWith(".jar") && !name.startsWith(module.getName());
                }
            });
            if (0 == bundles.length) {
                continue;
            }
            first.add(module);
        }
        sorted.addAll(first);
        final List<AbstractModule> toSort = new LinkedList<AbstractModule>(appModules);
        toSort.removeAll(first);
        final Stack<AbstractModule> moduleStack = new Stack<AbstractModule>();
        for (AbstractModule module : toSort) {
            visit(moduleStack, module, sorted);
        }
        return sorted;
    }

    private void visit(Stack<AbstractModule> moduleStack, AbstractModule module, List<AbstractModule> sorted) throws Exception {
        if (moduleStack.contains(module)) {
            throw new Exception("Cyclic dependency detected between modules: " + Tools.join(moduleStack, ",") + module.getName());
        }
        moduleStack.push(module);
        if (!sorted.contains(module)) {
            for (final Iterator<AbstractModule> dependentIt = module.getDependencies().iterator(); dependentIt.hasNext();) {
                visit(moduleStack, dependentIt.next(), sorted);
            }
            sorted.add(module);
        }
        moduleStack.pop();
    }
}
